
##############################################################################
#
# Copyright (c) 2005-2006 Balazs Ree <ree@greenfinity.hu> All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################


'''\
Log record
'''

import mechanize
from ClientCookie._HeadersUtil import split_header_words
from urllib import addinfourl
import httplib
from cStringIO import StringIO
from utils import lazy
from zExceptions import NotFound, BadRequest, InternalError, Forbidden, Redirect
## import mimetools

class ResponseHeaderWrapper(dict):
    '''Wrapper for the response headers. Originally this should be
    a mimetools Message object. However, we have the headers in
    a dict, so instead of producing it as a file and let mimetools
    parse it back, we attempt to emulate the used methods.

    If some methods become missing in a newer version, either they
    must be emulated as well or we can use the slower but more reliable
    Message generation.
    '''

    def getheaders(self, key):
        # this returns a list but we can have one occurance
        # anyway, since the data was in a dict
        try:
            return (self[key], )
        except KeyError:
            return ()

class ResponseWrapper(addinfourl):
    '''Wrapper for the response: like in urllib,
    but we need to provide the missing "seek" method.
    '''

    def __init__(self, fp, headers, url):
        addinfourl.__init__(self, fp, headers, url)
        # this seems to be missing
        self.seek = fp.seek
    
# XXX At the moment this is fixed.
_factory = mechanize.DefaultFactory()

class LogRecord(object):
    '''Represents one log entry
    
    Data can be logged via keywords at creation, 
    and later additional data can be logged into the 
    same logrecord via the log method.

    Logging of a logrecord has now two phases:
    
    1. logging of request data (LogRecord.__init__)
    2. logging of response data (LogRecord.log)
    '''

    
    def __init__(self, **kw):
        'Log the parameters initially'
#        if 'response_status' not in kw:
#            kw['response_status'] = None
#        if 'response_cookies' not in kw:
#            kw['response_cookies'] = None
        self.log(**kw)

    def log(self, **kw):
        'Log more elements into it (on response)'
        self.__dict__.update(kw)
        
    def _get_url(self):
        'Return the url of a logitem'
        return 'http://%s%s' % (self.environ['HTTP_HOST'], self.environ['PATH_INFO'])
            
    url = property(_get_url)

    # --
    # Response attributes
    # --

    def _get_wrapped_response(self):
        '''Returns a response wrapped the way urrlib would do

        or return None if there was no response
        '''
        try:
            response_body = self.response_body
            response_headers = self.response_headers
            response_url = self.url
        except (AttributeError, KeyError):
            # no response
            return None
        # the headers should be in the form of a mimetools Message
        # but we just use a wrapper shell to avoid double conversion
        ### headers_converted_back = ''.join(['%s: %s\n' % (k, v) for k, v in response_headers.iteritems()])
        ### headers_as_message = mimetools.Message(StringIO(headers_converted_back))
        headers_as_message = ResponseHeaderWrapper(response_headers)
        fp = StringIO(response_body)
        wrapped = ResponseWrapper(fp, headers_as_message, response_url)
        return wrapped
        
    wrapped_response = property(_get_wrapped_response)
    
    def _generate_response_forms(self):
        'Generate the response forms'
        response_forms = []
        response_links = []
        response = self.wrapped_response
        if response is None:
            response_forms_error = 'No response'
        else:
            if self.response_is_html:
                # attempt to parse it in as client forms
                try:
                    encoding = self._encoding(response)
                    response.seek(0)
                    response_forms = _factory.forms(response, encoding)
                    response.seek(0)
                    response_links = _factory.links(response, encoding)
                    response_forms_error = None
                except Exception, exc:
                    response_forms_error = 'Parsing error (%s: %s)' % (exc.__class__.__name__, exc)
            else:
                response_forms_error = 'Not html (content-type: %s)' % (self.response_content_type, )
        return response_forms, response_links, response_forms_error

    response_forms, response_links, response_forms_error = lazy(_generate_response_forms, 3)
     
    # XXX this is now a static value
    default_encoding = 'latin1'
    
    def _encoding(self, response):
        #
        # code extracted from mechanize.Browser._encoding
        #
        # HTTPEquivProcessor may be in use, so both HTTP and HTTP-EQUIV
        # headers may be in the response.  HTTP-EQUIV headers come last,
        # so try in order from first to last.
        for ct in response.info().getheaders("content-type"):
            for k, v in split_header_words([ct])[0]:
                if k == "charset":
                    return v
        return self.default_encoding

    def _response_content_type(self):
        try:
            return self.response_headers.get('content-type', '')
        except AttributeError:
            return ''

    response_content_type = property(_response_content_type)
    
    response_is_html = property(lambda self: self.response_content_type.startswith('text/html'))
    response_is_redirect = property(lambda self: self.response_status is not None and 
                    self.response_status / 100 == 3 or
                    self.publish_exc[0] is Redirect or self.publish_exc[0] == 'Redirect')
                    # redirect = status 30x or Redirect exception
    response_is_unauthorized = property(lambda self: self.publish_exc[0] == 'Unauthorized')

    def _get_response_redirect_url(self):
        'Gets the redirect location url, without the parameters.'
        # there are two cases: either redirect with response, or via exception.
        if self.publish_exc[0] is Redirect or self.publish_exc[0] == 'Redirect':
            # Redirect exception
            url = str(self.publish_exc[1])
        else:
            # status 30x
            url = getattr(self, 'response_headers', {}).get('location', '')
        split = httplib.urlsplit(url)[:3]
        if split[0]:
            split = split[:1] + ('://', ) + split[1:]
        url_without_parms = ''.join(split)
        return url_without_parms
        
    response_redirect_url = property(_get_response_redirect_url) 


